# inspired from:

# https://python-packaging-user-guide.readthedocs.org/en/latest/single_source_version/
# https://github.com/pypa/warehouse/blob/master/warehouse/__about__.py

__name__ = 'thinc'
__version__ = '6.10.2'
__summary__ = "Practical Machine Learning for NLP"
__uri__ = 'https://github.com/explosion/thinc'
__author__ = 'Matthew Honnibal'
__email__ = 'matt@explosion.ai'
__license__ = 'MIT'
__title__ = "thinc"
__release__ = True



from cymem.cymem cimport Pool
from preshed.maps cimport PreshMap

from .features cimport ConjunctionExtracter
from ..typedefs cimport weight_t, feat_t, class_t
from ..structs cimport FeatureC
from ..structs cimport ExampleC


cdef class AveragedPerceptron:
    cdef readonly Pool mem
    cdef readonly PreshMap weights
    cdef readonly PreshMap averages
    cdef readonly PreshMap lasso_ledger
    cdef ConjunctionExtracter extracter
    cdef public int time
    cdef public weight_t learn_rate
    cdef public weight_t l1_penalty
    cdef public weight_t momentum

    cdef void set_scoresC(self, weight_t* scores, const FeatureC* feats, int nr_feat) nogil
    cdef int updateC(self, const ExampleC* eg) except -1
    cpdef int update_weight(self, feat_t feat_id, class_t clas, weight_t upd) except -1
    cpdef int update_weight_ftrl(self, feat_t feat_id, class_t clas, weight_t upd) except -1



from __future__ import print_function, unicode_literals
import numpy
from preshed.maps import PreshMap
from pathlib import Path

try:
    import cupy
    from cupy import get_array_module
except ImportError:
    cupy = None
    get_array_module = lambda _: numpy


def get_ops(ops):
    from .ops import NumpyOps, CupyOps
    if ops in ('numpy', 'cpu'):
        return NumpyOps
    elif ops in ('cupy', 'gpu'):
        return CupyOps
    else:
        raise ValueError("TODO error %s" % ops)


def mark_sentence_boundaries(sequences, drop=0.): # pragma: no cover
    '''Pad sentence sequences with EOL markers.'''
    for sequence in sequences:
        sequence.insert(0, '-EOL-')
        sequence.insert(0, '-EOL-')
        sequence.append('-EOL-')
        sequence.append('-EOL-')
    return sequences, None


def remap_ids(ops):
    id_map = {0: 0}
    def begin_update(ids, drop=0.):
        n_vector = len(id_map)
        for i, id_ in enumerate(ids):
            if id_ not in id_map:
                id_map[id_] = n_vector
                n_vector += 1
            ids[i] = id_map[id_]
        return ids, None
    return begin_update

def copy_array(dst, src, casting='same_kind', where=None):
    if isinstance(dst, numpy.ndarray) and isinstance(src, numpy.ndarray):
        dst[:] = src
    elif isinstance(dst, cupy.ndarray):
        src = cupy.array(src, copy=False)
        cupy.copyto(dst, src)
    else:
        numpy.copyto(dst, src)


def ensure_path(path):
    if isinstance(path, basestring) or isinstance(path, str):
        return Path(path)
    else:
        return path


def to_categorical(y, nb_classes=None):
    # From keras
    xp = get_array_module(y)
    if xp is cupy:
        y = y.get()
    y = numpy.array(y, dtype='int').ravel()
    if not nb_classes:
        nb_classes = numpy.max(y) + 1
    n = y.shape[0]
    categorical = numpy.zeros((n, nb_classes), dtype='float32')
    categorical[numpy.arange(n), y] = 1
    return xp.asarray(categorical)


def flatten_sequences(sequences, drop=0.): # pragma: no cover
    xp = get_array_module(sequences[0])
    return xp.concatenate(sequences), None


def partition(examples, split_size): # pragma: no cover
    examples = list(examples)
    numpy.random.shuffle(examples)
    n_docs = len(examples)
    split = int(n_docs * split_size)
    return examples[:split], examples[split:]


def minibatch(stream, batch_size=1000): # pragma: no cover
    if hasattr(stream, '__len__') and hasattr(stream, '__getitem__'):
        i = 0
        while i < len(stream):
            yield stream[i : i + batch_size]
            i += batch_size
    else:
        batch = []
        for X in stream:
            batch.append(X)
            if len(batch) >= batch_size:
                yield batch
                batch = []
        if len(batch) != 0:
            yield batch



import pytest
import numpy

from ...neural._classes.affine import Affine
from ...exceptions import UndefinedOperatorError, DifferentLengthError
from ...exceptions import ExpectedTypeError, ShapeMismatchError


def test_mismatched_shapes_raise_ShapeError():
    X = numpy.ones((3, 4))
    model = Affine(10, 5)
    with pytest.raises(ShapeMismatchError):
        y = model.predict(X)




from libc.stdint cimport uint32_t, uint64_t

cimport numpy as np
import numpy as np


assert sizeof(int) == sizeof(np.int32_t)


cdef extern from "_cuda_shim.h":
    void gpu_maxout(float* best__bo, int* which__bo,
        const float* cands__bop, int B, int O, int P)
    void gpu_mean_pool(float* means,
        const float* X, const int* lengths, int B, int T, int O) nogil
    void gpu_sum_pool(float* sums,
        const float* X, const int* lengths, int B, int T, int O) nogil
    void gpu_max_pool(float* maxes, int* which,
        const float* X, const int* lengths, int B, int T, int O) nogil
    void gpu_backprop_mean_pool(float* dX,
        const float* d_means, const int* lengths, int B, int T, int O) nogil
    void gpu_backprop_sum_pool(float* dX,
        const float* d_sums, const int* lengths, int B, int T, int O) nogil
    void gpu_backprop_max_pool(float* dX,
        const float* d_maxes, const int* which, const int* lengths, int B, int T, int O) nogil
    void gpu_hash_data(char* dest,
        const char* src, size_t out_size, size_t in_size, size_t n_items, uint32_t seed) nogil


def maxout(*args, **kwargs):
    pass

def mean_pool(ops, x, lengths):
    means = ops.allocate((lengths.shape[0], x.shape[1]))
    cdef size_t means_ptr = means.data.ptr
    cdef size_t x_ptr = x.data.ptr
    cdef size_t lengths_ptr = lengths.data.ptr
    gpu_mean_pool(<float*>means_ptr,
        <const float*>x_ptr, <const int*>lengths_ptr, lengths.shape[0], x.shape[0], x.shape[1])
    return means


def max_pool(ops, x, lengths):
    maxes = ops.allocate((lengths.shape[0], x.shape[1]))
    which = ops.allocate((lengths.shape[0], x.shape[1]), dtype='i')
    cdef size_t maxes_ptr = maxes.data.ptr
    cdef size_t which_ptr = which.data.ptr
    cdef size_t x_ptr = x.data.ptr
    cdef size_t lengths_ptr = lengths.data.ptr
    gpu_max_pool(<float*>maxes_ptr, <int*>which_ptr,
        <const float*>x_ptr, <const int*>lengths_ptr, lengths.shape[0], x.shape[0], x.shape[1])
    return maxes, which


def sum_pool(ops, x, lengths):
    sums = ops.allocate((lengths.shape[0], x.shape[1]))
    cdef size_t sums_ptr = sums.data.ptr
    cdef size_t x_ptr = x.data.ptr
    cdef size_t lengths_ptr = lengths.data.ptr
    gpu_sum_pool(<float*>sums_ptr,
        <const float*>x_ptr, <const int*>lengths_ptr, lengths.shape[0], x.shape[0], x.shape[1])
    return sums


def backprop_mean_pool(ops, d_means, lengths):
    dX = ops.allocate((lengths.sum().get(), d_means.shape[1]))
    cdef size_t d_means_ptr = d_means.data.ptr
    cdef size_t dX_ptr = dX.data.ptr
    cdef size_t lengths_ptr = lengths.data.ptr
    cdef int B = lengths.shape[0]
    cdef int T = dX.shape[0]
    cdef int O = dX.shape[1]

    gpu_backprop_mean_pool(<float*>dX_ptr,
        <const float*>d_means_ptr, <const int*>lengths_ptr,
        B, T, O)
    return dX


def backprop_sum_pool(ops, d_sums, lengths):
    dX = ops.allocate((lengths.sum().get(), d_sums.shape[1]))
    cdef size_t d_sums_ptr = d_sums.data.ptr
    cdef size_t dX_ptr = dX.data.ptr
    cdef size_t lengths_ptr = lengths.data.ptr
    cdef int B = lengths.shape[0]
    cdef int T = dX.shape[0]
    cdef int O = dX.shape[1]

    gpu_backprop_sum_pool(<float*>dX_ptr,
        <const float*>d_sums_ptr, <const int*>lengths_ptr,
        B, T, O)
    return dX

def backprop_max_pool(ops, d_maxes, which, lengths):
    dX = ops.allocate((lengths.sum().get(), d_maxes.shape[1]))
    cdef size_t d_maxes_ptr = d_maxes.data.ptr
    cdef size_t which_ptr = which.data.ptr
    cdef size_t dX_ptr = dX.data.ptr
    cdef size_t lengths_ptr = lengths.data.ptr
    cdef int B = lengths.shape[0]
    cdef int T = dX.shape[0]
    cdef int O = dX.shape[1]

    gpu_backprop_max_pool(<float*>dX_ptr,
        <const float*>d_maxes_ptr, <const int*>which_ptr, <const int*>lengths_ptr,
        B, T, O)
    return dX


def hash(ops, ids, seed):
    keys = ops.allocate((ids.shape[0], 4), dtype='uint32')
    cdef size_t keys_ptr = keys.data.ptr
    cdef size_t ids_ptr = ids.data.ptr

    gpu_hash_data(<char*>keys_ptr,
        <const char*>ids_ptr, sizeof(uint32_t)*4, sizeof(uint64_t), ids.shape[0], seed)
    return keys



from cymem.cymem cimport Pool
from libc.math cimport sqrt as c_sqrt
from libc.string cimport memset, memcpy, memmove

from preshed.maps cimport map_init as Map_init
from preshed.maps cimport map_set as Map_set
from preshed.maps cimport map_get as Map_get

from ..structs cimport ExampleC, FeatureC, MapC
from ..typedefs cimport feat_t, weight_t, atom_t
from ..linalg cimport Vec, VecVec


cdef class Example:
    cdef Pool mem
    cdef ExampleC c

    @staticmethod
    cdef inline Example from_ptr(ExampleC* ptr):
        cdef Example eg = Example.__new__(Example)
        eg.c = ptr[0]
        return eg
